#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <iostream>
#include <list>
#include <glibmm/miscutils.h>
#include <glibmm/spawn.h>
#include <glibmm/thread.h>

#include "converter.hpp"
#include "info.hpp"

// External variables, both are from main.cpp
extern bool verbose;
extern std::string sep;

int 		current_id;
std::string	tmp_file;
std::string	current_output_path;

// pipes
int ffin;
int ffout;
int fferr;
int fds_info[2];	// info sent to parent process

Glib::Pid ffmpeg_pid; // this is ffmpeg's pid

// Kill the child process (ffmpeg) and exit
void signal_handler (int signal_number)
{

	if(ffmpeg_pid)
	{
		kill (ffmpeg_pid, SIGTERM);
		Glib::spawn_close_pid (ffmpeg_pid);
	}

	unlink (tmp_file.c_str());

	abort();

}

// calculate total padding required
unsigned int calc_padding (input_t input, output_t output)
{

	unsigned int padding = 0;
	unsigned int input_aspect	= input.aspect * 10;
	unsigned int output_aspect	= (atof (output.aspectw.c_str()) / atof (output.aspecth.c_str()) * 10);
	double output_width		= atof ((output.sizew).c_str());
	double output_height		= atof ((output.sizeh).c_str());


	if(input_aspect > output_aspect)
	{
		if(verbose)
			std::cout << sep << "input & output aspects doesn't match:\nInput aspect:\t" << input_aspect
				  << "\nOutput aspect:\t" << output_aspect << "\ncalculatting pad...";

		padding = ( output_height - input.sizeh / (input.sizew/output_width) );
	}

	return padding;

}

int convert (std::list< io_t > io_list)
{

	// Handle SIGTERM to kill the child process (ffmpeg), otherwise it loses it's parent process & continues running out off our control :)
	// for more info on signal handling, see Advanced Linux Programming 3.3 Signals page 54
	struct sigaction sa;
	memset( &sa, 0, sizeof( sa ) );
	sa.sa_handler = &signal_handler;
	sigaction( SIGTERM, &sa, NULL );

	int child_status;
	pid_t child_pid;

	pipe (fds_info); // create a pipe

	unsigned int i;

	child_pid = fork();
	if( child_pid != 0)
	{
		return child_pid; // this is the parent process, return so we can continue running
	}
	else
	{
		// these are used when reconverting
		// we convert a file to a temporary file then reconvert the temporary file to final file
		// 'Creative Zen V Plus' needs this or else we get an vertically flipped video
		bool reconverting = false;
		bool rmtmp = false;
		io_t re_io;

		int i = 1;
		std::list< io_t >::iterator iter = io_list.begin();
		while(iter != io_list.end())
		{
			io_t io;
			if(reconverting)
				io = re_io;
			else
				io = *iter;

			unsigned int padding = calc_padding (io.input, io.output);
			// frame size should be a multiple of 2
			if(padding%2)
				padding -= 1;
			// top and bottom padding should be a multiple of 2
			if(padding/2%2)
				padding += 2;

			char char_padding[5], char_size[10];
			sprintf (char_padding, "%d", padding/2);
			// size = output frame size - padding
			sprintf (char_size, "%sx%d", io.output.sizew.c_str(), atoi(io.output.sizeh.c_str())-padding);

			if(verbose && padding)
				std::cout << "\ntotal padding:\t" << padding << "\ntop pad:\t" << padding/2 << "\nbottom pad:\t" << padding/2 << "\nOutput size - padding = " << char_size << std::endl;

			std::list<std::string> arglist;

			arglist.push_back ("ffmpeg");

			arglist.push_back ("-i");		arglist.push_back (io.input.path);

			// overwrite output file?
			if(io.output.overwrite)
				arglist.push_back ("-y");

			arglist.push_back ("-s");		arglist.push_back (char_size);
			arglist.push_back ("-aspect");		arglist.push_back (io.output.aspectw + ":" + io.output.aspecth);
			arglist.push_back ("-vcodec");		arglist.push_back (io.output.vcodec);
			arglist.push_back ("-b");		arglist.push_back (io.output.vbitrate);
			arglist.push_back ("-r");		arglist.push_back (io.output.framerate);

			// use pix_fmt?
			if(io.output.usepix_fmt)
			{
				arglist.push_back ("-pix_fmt");	arglist.push_back (io.output.pix_fmt);
			}

			arglist.push_back ("-deinterlace");
			arglist.push_back ("-padtop");		arglist.push_back (char_padding);
			arglist.push_back ("-padbottom");	arglist.push_back (char_padding);

			arglist.push_back ("-acodec");		arglist.push_back (io.output.acodec);
			arglist.push_back ("-ab");		arglist.push_back (io.output.abitrate);
			arglist.push_back ("-ar");		arglist.push_back (io.output.samplerate);
			arglist.push_back ("-ac");		arglist.push_back (io.output.achannels);

			// use flags?
			if(io.output.useflags)
			{
				arglist.push_back ("-mbd");	arglist.push_back (io.output.mbd);
				arglist.push_back ("-flags");	arglist.push_back (io.output.flags);
				arglist.push_back ("-cmp");	arglist.push_back (io.output.cmp);
				arglist.push_back ("-subcmp");	arglist.push_back (io.output.subcmp);
				arglist.push_back ("-g");	arglist.push_back (io.output.g);
			}

			// use reconverting?
			if(io.output.reconvert)
			{
				tmp_file.assign (Glib::path_get_dirname (io.output.path));
				tmp_file.append ("/tmp");
				tmp_file.append (io.output.path, io.output.path.rfind("."), io.output.path.size());

				re_io = io;
				re_io.input.path = tmp_file;
				re_io.output.reconvert = false;
				
				io.output.path = tmp_file;

			}

			arglist.push_back (io.output.path);

			if(verbose)
				std::cout << sep << "Converting...\nInput:\t'" << io.input.path << "'\nOutput:\t'" << io.output.path << "'" << sep;

			try
			{
				Glib::spawn_async_with_pipes	(Glib::get_home_dir(),
								arglist,
								Glib::SPAWN_DO_NOT_REAP_CHILD|Glib::SPAWN_SEARCH_PATH,
								sigc::slot< void >(),
								&ffmpeg_pid,
								&ffin, &ffout, &fferr);
			}
			catch(const Glib::SpawnError& error)
			{

				std::cout << sep << error.what() << std::endl;

			}

			current_id = i;
			current_output_path = io.output.path;

			Glib::Thread::create (sigc::ptr_fun (&put_info), false);

			wait (&child_status); // Wait for each child process to finish

			Glib::spawn_close_pid (ffmpeg_pid);

			if(reconverting)
			{
				reconverting = false;

				if(verbose)
					std::cout << sep << "Removing temporary file: '" << tmp_file << "'" << std::endl;

				unlink (tmp_file.c_str());
			}

			if(io.output.reconvert)
				reconverting = true;

			else
			{
				i++;
				iter++;
			}
		}
		close (fds_info[1]);
	}

	// If we use return here we will return to the parent,
	// but we did it when we return the child pid,
	// so there will be two processes exactly the same causing strange behavior from gui & printing Xlib errors.
	abort();

}
